@@ -5,14 +5,14 @@ import sys
 import urllib.parse as urlparse
 import uuid
 from pathlib import Path
-from typing import Any, Dict, List, Mapping, Optional, Tuple, Union
+from typing import Any, Dict, Iterator, List, Mapping, Optional, Sequence, Tuple, Union
 
 import async_retriever as ar
 import cytoolz as tlz
 import pyproj
 from owslib.wfs import WebFeatureService
 from owslib.wms import WebMapService
-from pydantic import AnyHttpUrl, BaseModel, validator
+from pydantic import BaseModel, validator
 from shapely.geometry import LineString, MultiPoint, Point, Polygon
 
 from . import utils
@@ -35,6 +35,26 @@ DEF_CRS = "epsg:4326"
 EXPIRE = -1
 
 
+def validate_version(val: str, valid_versions: List[str]) -> str:
+    """Validate version from a list of valid versions.
+
+    Parameters
+    ----------
+    val : str
+        Input version value.
+    valid_versions : list of str
+        List of valid versions.
+
+    Returns
+    -------
+    str
+        Validated version value.
+    """
+    if val not in valid_versions:
+        raise InvalidInputValue("version", valid_versions)
+    return val
+
+
 class RESTValidator(BaseModel):
     """Validate ArcGISRESTful inputs.
 
@@ -73,7 +93,7 @@ class RESTValidator(BaseModel):
         If ``True``, disable caching requests, defaults to False.
     """
 
-    base_url: AnyHttpUrl
+    base_url: str
     layer: Optional[int] = None
     outformat: str = "geojson"
     outfields: Union[List[str], str] = "*"
@@ -85,7 +105,7 @@ class RESTValidator(BaseModel):
     disable_caching: bool = False
 
     @validator("base_url")
-    def _layer_from_url(cls, v: AnyHttpUrl) -> Tuple[str, Optional[int]]:
+    def _layer_from_url(cls, v: str) -> Tuple[str, Optional[int]]:
         return cls._split_url(urlparse.unquote(v))
 
     @validator("layer")
@@ -103,11 +123,8 @@ class RESTValidator(BaseModel):
         return v if isinstance(v, list) else [v]
 
     @validator("crs")
-    def _valid_crs(cls, v: str) -> pyproj.CRS:
-        try:
-            return pyproj.CRS(v)
-        except pyproj.exceptions.CRSError as ex:
-            raise InvalidInputType("crs", "a valid CRS") from ex
+    def _valid_crs(cls, v: str) -> str:
+        return utils.validate_crs(v)
 
     @validator("max_workers")
     def _positive_integer_threads(cls, v: int) -> int:
@@ -169,11 +186,11 @@ class ArcGISRESTfulBase:
 
     def __init__(
         self,
-        base_url: AnyHttpUrl,
+        base_url: str,
         layer: Optional[int] = None,
         outformat: str = "geojson",
         outfields: Union[List[str], str] = "*",
-        crs: Union[str, pyproj.CRS] = DEF_CRS,
+        crs: str = DEF_CRS,
         max_workers: int = 1,
         verbose: bool = False,
         disable_retry: bool = False,
@@ -219,7 +236,7 @@ class ArcGISRESTfulBase:
         self.return_geom: bool = True
         self.n_missing: int = 0
         self.total_n_features: int = 0
-        self.failed_path: Optional[Path] = None
+        self.failed_path: Union[str, Path] = ""
         self.request_id: Optional[str] = None
 
         self.initialize_service()
@@ -239,7 +256,7 @@ class ArcGISRESTfulBase:
         if any(f not in self.valid_fields for f in self.outfields):
             raise InvalidInputValue("outfields", self.valid_fields)
 
-    def partition_oids(self, oids: Union[List[int], int]) -> List[Tuple[str, ...]]:
+    def partition_oids(self, oids: Union[List[int], int]) -> Iterator[Tuple[str, ...]]:
         """Partition feature IDs based on ``self.max_nrecords``."""
         oid_list = [oids] if isinstance(oids, int) else set(oids)
         if len(oid_list) == 0:
@@ -252,7 +269,7 @@ class ArcGISRESTfulBase:
 
     def get_features(
         self,
-        featureids: List[Tuple[str, ...]],
+        featureids: Iterator[Tuple[str, ...]],
         return_m: bool = False,
         return_geom: bool = True,
     ) -> List[Dict[str, Any]]:
@@ -277,7 +294,7 @@ class ArcGISRESTfulBase:
             {
                 "objectIds": ",".join(ids),
                 "returnGeometry": f"{return_geom}".lower(),
-                "outSR": self.out_sr,
+                "outSR": f"{self.out_sr}",
                 "outfields": ",".join(self.outfields),
                 "ReturnM": f"{return_m}".lower(),
                 "f": self.outformat,
@@ -295,22 +312,27 @@ class ArcGISRESTfulBase:
         return resp
 
     def _set_service_properties(self) -> None:
+        rjson = self.get_response(self.base_url, [{"f": "json"}])[0]
         try:
-            rjson = self.get_response(self.base_url, [{"f": "json"}])[0]
             self.valid_layers = {f"{lyr['id']}": lyr["name"] for lyr in rjson["layers"]}
             self.query_formats = rjson["supportedQueryFormats"].replace(" ", "").lower().split(",")
 
             extent = rjson["extent"] if "extent" in rjson else rjson["fullExtent"]
             bounds = (extent["xmin"], extent["ymin"], extent["xmax"], extent["ymax"])
-            for f in ["latestWkid", "wkid", "wkt"]:
-                if f in extent["spatialReference"]:
-                    crs = extent["spatialReference"][f]
-                    break
-            self.extent = utils.match_crs(bounds, crs, DEF_CRS)
-
         except (ValueError, KeyError) as ex:
             raise ServiceError(self.base_url) from ex
 
+        crs_iter = iter(["latestWkid", "wkid", "wkt"])
+        while True:
+            try:
+                crs = utils.validate_crs(extent["spatialReference"].get(next(crs_iter)))
+                self.extent = utils.match_crs(bounds, crs, DEF_CRS)
+                break
+            except InvalidInputType:
+                continue
+            except StopIteration as ex:
+                raise ServiceError(self.base_url) from ex
+
         with contextlib.suppress(KeyError):
             self.units = rjson["units"].replace("esri", "").lower()
             self.max_nrecords = int(rjson["maxRecordCount"])
@@ -343,7 +365,7 @@ class ArcGISRESTfulBase:
             MultiPoint,
             Tuple[float, float, float, float],
         ],
-        geo_crs: str = DEF_CRS,
+        geo_crs: Union[str, pyproj.CRS] = DEF_CRS,
     ) -> Mapping[str, str]:
         """Generate geometry queries based on ESRI template."""
         geom = utils.match_crs(geom, geo_crs, self.crs)
@@ -369,7 +391,7 @@ class ArcGISRESTfulBase:
         self, return_m: bool, return_geo: bool, partition_fac: float
     ) -> List[Dict[str, Any]]:
         """Retry failed requests."""
-        with open(self.failed_path) as f:  # type: ignore
+        with open(self.failed_path) as f:
             oids = [int(i) for i in f.read().splitlines()]
 
         max_nrecords = self.max_nrecords
@@ -401,16 +423,14 @@ class ArcGISRESTfulBase:
         return list(tlz.concat(features))
 
     def get_response(
-        self, url: str, payloads: List[Dict[str, str]], method: str = "GET"
+        self, url: str, payloads: Sequence[Dict[str, str]], method: str = "GET"
     ) -> List[Dict[str, Any]]:
         """Send payload and get the response."""
         req_key = "params" if method == "GET" else "data"
-        urls, kwds = zip(*((url, {req_key: p}) for p in payloads))
         try:
-            return ar.retrieve(  # type: ignore
-                urls,
-                "json",
-                list(kwds),
+            return ar.retrieve_json(
+                [url] * len(payloads),
+                [{req_key: p} for p in payloads],
                 request_method=method,
                 max_workers=self.max_workers,
                 disable=self.disable_caching,
@@ -420,7 +440,7 @@ class ArcGISRESTfulBase:
             raise ZeroMatched
 
     def _cleanup_resp(
-        self, resp: List[Dict[str, Any]], payloads: List[Dict[str, str]]
+        self, resp: List[Dict[str, Any]], payloads: Sequence[Dict[str, str]]
     ) -> List[Dict[str, Any]]:
         """Remove failed responses."""
         fails = [i for i, r in enumerate(resp) if "error" in r]
@@ -443,24 +463,22 @@ class ArcGISRESTfulBase:
                     self.return_m = bool(payloads[0]["ReturnM"])
                     self.return_geom = bool(payloads[0]["returnGeometry"])
                     self.total_n_features = self.n_features
-                    if self.verbose:
-                        logger.info(f"Found {len(fails)} failed requests. Retrying ...")
+                    logger.warning(f"Found {len(fails)} failed requests. Retrying ...")
                     resp.extend(self.retry_failed_requests())
 
-                    if self.verbose:
-                        logger.warning(
-                            " ".join(
-                                [
-                                    f"Total of {self.n_missing} out of {self.total_n_features}",
-                                    "requested features are not available in the dataset.",
-                                    "Returning the successfully retrieved features."
-                                    "The failed object IDs have been saved in the",
-                                    f"file {self.failed_path}. The service returned the",
-                                    "following error message for the failed requests:\n",
-                                    err,
-                                ]
-                            )
+                    logger.warning(
+                        " ".join(
+                            [
+                                f"Total of {self.n_missing} out of {self.total_n_features}",
+                                "requested features are not available in the dataset.",
+                                "Returning the successfully retrieved features.",
+                                "The failed object IDs have been saved in the",
+                                f"file {self.failed_path}. The service returned the",
+                                "following error message for the failed requests:\n",
+                                err,
+                            ]
                         )
+                    )
 
         return resp
 
@@ -499,14 +517,14 @@ class WMSBase(BaseModel):
         The WMS service version which should be either 1.1.1 or 1.3.0, defaults to 1.3.0.
     crs : str, optional
         The spatial reference system to be used for requesting the data, defaults to
-        epsg:4326.
+        ``epsg:4326``.
     expire_after : int, optional
         Expiration time for response caching in seconds, defaults to -1 (never expire).
     disable_caching : bool, optional
         If ``True``, disable caching requests, defaults to False.
     """
 
-    url: AnyHttpUrl
+    url: str
     layers: Union[str, List[str]]
     outformat: str
     version: str = "1.3.0"
@@ -514,6 +532,14 @@ class WMSBase(BaseModel):
     expire_after: float = EXPIRE
     disable_caching: bool = False
 
+    @validator("crs")
+    def _valid_crs(cls, v: str) -> str:
+        return utils.validate_crs(v)
+
+    @validator("version")
+    def _version(cls, v: str) -> str:
+        return validate_version(v, ["1.1.1", "1.3.0"])
+
     def __repr__(self) -> str:
         """Print the services properties."""
         layers = self.layers if isinstance(self.layers, list) else [self.layers]
@@ -543,7 +569,7 @@ class WMSBase(BaseModel):
             raise InvalidInputValue("outformat", valid_outformats)
 
         valid_crss = {lyr: [s.lower() for s in wms[lyr].crsOptions] for lyr in layers}
-        if any(self.crs not in valid_crss[lyr] for lyr in layers):
+        if any(self.crs.lower() not in valid_crss[lyr] for lyr in layers):
             _valid_crss = (f"{lyr}: {', '.join(cs)}\n" for lyr, cs in valid_crss.items())
             raise InvalidInputValue("CRS", _valid_crss)
 
@@ -572,25 +598,25 @@ class WFSBase(BaseModel):
         The data format to request for data from the service, defaults to None which
          throws an error and includes all the available format offered by the service.
     version : str, optional
-        The WFS service version which should be either 1.0.0, 1.1.0, or 2.0.0.
-        Defaults to 2.0.0.
+        The WFS service version which should be either ``1.0.0``, ``1.1.0``, or
+        ``2.0.0``. Defaults to ``2.0.0``.
     crs: str, optional
         The spatial reference system to be used for requesting the data, defaults to
-        epsg:4326.
+        ``epsg:4326``.
     read_method : str, optional
         Method for reading the retrieved data, defaults to ``json``. Valid options are
         ``json``, ``binary``, and ``text``.
     max_nrecords : int, optional
         The maximum number of records in a single request to be retrieved from the service,
-        defaults to 1000. If the number of records requested is greater than this value,
-        it will be split into multiple requests.
+        defaults to 1000. If the number of requested records is greater than this value,
+        the query will be split into multiple requests.
     expire_after : int, optional
         Expiration time for response caching in seconds, defaults to -1 (never expire).
     disable_caching : bool, optional
-        If ``True``, disable caching requests, defaults to False.
+        If ``True``, disable caching requests, defaults to ``False``.
     """
 
-    url: AnyHttpUrl
+    url: str
     layer: Optional[str] = None
     outformat: Optional[str] = None
     version: str = "2.0.0"
@@ -600,6 +626,21 @@ class WFSBase(BaseModel):
     expire_after: float = EXPIRE
     disable_caching: bool = False
 
+    @validator("read_method")
+    def _read_method(cls, v: str) -> str:
+        valid_methods = ["json", "binary", "text"]
+        if v not in valid_methods:
+            raise InvalidInputValue("read_method", valid_methods)
+        return v
+
+    @validator("crs")
+    def _valid_crs(cls, v: str) -> str:
+        return utils.validate_crs(v)
+
+    @validator("version")
+    def _version(cls, v: str) -> str:
+        return validate_version(v, ["1.0.0", "1.1.0", "2.0.0"])
+
     def __repr__(self) -> str:
         """Print the services properties."""
         return "\n".join(
@@ -664,22 +705,21 @@ class WFSBase(BaseModel):
             "typeName": self.layer,
             max_features: 1,
         }
-        rjson = ar.retrieve(
+        rjson = ar.retrieve_json(
             [self.url],
-            "json",
             [{"params": payload}],
             expire_after=self.expire_after,
             disable=self.disable_caching,
         )[0]
         valid_fields = list(
             set(
-                utils.traverse_json(rjson, ["fields", "name"])  # type: ignore
-                + utils.traverse_json(rjson, ["fields", "alias"])  # type: ignore
+                utils.traverse_json(rjson, ["fields", "name"])
+                + utils.traverse_json(rjson, ["fields", "alias"])
                 + ["*"]
             )
         )
 
         if None in valid_fields:
-            valid_fields = list(utils.traverse_json(rjson, ["features", "properties"])[0].keys())  # type: ignore
+            valid_fields = list(utils.traverse_json(rjson, ["features", "properties"])[0].keys())
 
         return valid_fields
