@@ -1,8 +1,10 @@
 """Customized PyGeoOGC exceptions."""
-from typing import Generator, List, Optional, Union
+from typing import Optional
+
+import async_retriever as ar
 
 
-class ServiceError(Exception):
+class ServiceError(ar.ServiceError):
     """Exception raised when the requested data is not available on the server.
 
     Parameters
@@ -11,12 +13,35 @@ class ServiceError(Exception):
         Service error message.
     """
 
-    def __init__(self, err: str) -> None:
-        self.message = f"Service returned the following error message:\n{err}"
-        super().__init__(self.message)
 
-    def __str__(self) -> str:
-        return self.message
+class InvalidInputValue(ar.InvalidInputValue):
+    """Exception raised for invalid input.
+
+    Parameters
+    ----------
+    inp : str
+        Name of the input parameter
+    valid_inputs : tuple
+        List of valid inputs
+    """
+
+
+class InvalidInputType(ar.InvalidInputType):
+    """Exception raised when a function argument type is invalid.
+
+    Parameters
+    ----------
+    arg : str
+        Name of the function argument
+    valid_type : str
+        The valid type of the argument
+    example : str, optional
+        An example of a valid form of the argument, defaults to None.
+    """
+
+
+class MissingInputs(ValueError):
+    """Exception raised when there are missing function arguments."""
 
 
 class ServiceUnavailable(Exception):
@@ -54,53 +79,3 @@ class ZeroMatched(ValueError):
 
     def __str__(self) -> str:
         return self.message
-
-
-class InvalidInputValue(Exception):
-    """Exception raised for invalid input.
-
-    Parameters
-    ----------
-    inp : str
-        Name of the input parameter
-    valid_inputs : tuple
-        List of valid inputs
-    """
-
-    def __init__(
-        self, inp: str, valid_inputs: Union[List[str], Generator[str, None, None]]
-    ) -> None:
-        self.message = f"Given {inp} is invalid. Valid options are:\n" + "\n".join(
-            str(i) for i in valid_inputs
-        )
-        super().__init__(self.message)
-
-    def __str__(self) -> str:
-        return self.message
-
-
-class InvalidInputType(Exception):
-    """Exception raised when a function argument type is invalid.
-
-    Parameters
-    ----------
-    arg : str
-        Name of the function argument
-    valid_type : str
-        The valid type of the argument
-    example : str, optional
-        An example of a valid form of the argument, defaults to None.
-    """
-
-    def __init__(self, arg: str, valid_type: str, example: Optional[str] = None) -> None:
-        self.message = f"The {arg} argument should be of type {valid_type}"
-        if example is not None:
-            self.message += f":\n{example}"
-        super().__init__(self.message)
-
-    def __str__(self) -> str:
-        return self.message
-
-
-class MissingInputs(ValueError):
-    """Exception raised when there are missing function arguments."""
