@@ -3,14 +3,13 @@ import itertools
 import uuid
 from pathlib import Path
 from types import SimpleNamespace
-from typing import Any, Dict, List, Optional, Tuple, Union
+from typing import Any, Dict, Iterator, List, Optional, Tuple, Union
 
 import async_retriever as ar
 import cytoolz as tlz
 import pyproj
 import shapely.ops as ops
 import yaml
-from pydantic import AnyHttpUrl
 from shapely.geometry import LineString, MultiPoint, MultiPolygon, Point, Polygon
 
 from . import utils
@@ -44,7 +43,7 @@ class ArcGISRESTful:
         The output fields to be requested. Setting ``*`` as outfields requests
         all the available fields which is the default behaviour.
     crs : str, optional
-        The spatial reference of the output data, defaults to EPSG:4326
+        The spatial reference of the output data, defaults to ``epsg:4326``.
     max_workers : int, optional
         Number of simultaneous download, default to 1, i.e., no threading. Note
         that some services might face issues when several requests are sent
@@ -66,11 +65,11 @@ class ArcGISRESTful:
 
     def __init__(
         self,
-        base_url: AnyHttpUrl,
+        base_url: str,
         layer: Optional[int] = None,
         outformat: str = "geojson",
         outfields: Union[List[str], str] = "*",
-        crs: Union[str, pyproj.CRS] = DEF_CRS,
+        crs: str = DEF_CRS,
         max_workers: int = 1,
         verbose: bool = False,
         disable_retry: bool = False,
@@ -101,11 +100,11 @@ class ArcGISRESTful:
             List[Tuple[float, float]],
             Tuple[float, float, float, float],
         ],
-        geo_crs: str = DEF_CRS,
+        geo_crs: Union[str, pyproj.CRS] = DEF_CRS,
         spatial_relation: str = "esriSpatialRelIntersects",
         sql_clause: Optional[str] = None,
         distance: Optional[int] = None,
-    ) -> List[Tuple[str, ...]]:
+    ) -> Iterator[Tuple[str, ...]]:
         """Get feature IDs within a geometry that can be combined with a SQL where clause.
 
         Parameters
@@ -114,8 +113,8 @@ class ArcGISRESTful:
             A geometry (LineString, Polygon, Point, MultiPoint), tuple of length two
             (``(x, y)``), a list of tuples of length 2 (``[(x, y), ...]``), or bounding box
             (tuple of length 4 (``(xmin, ymin, xmax, ymax)``)).
-        geo_crs : str
-            The spatial reference of the input geometry, defaults to EPSG:4326.
+        geo_crs : str or pyproj.CRS
+            The spatial reference of the input geometry.
         spatial_relation : str, optional
             The spatial relationship to be applied on the input geometry
             while performing the query. If not correct a list of available options is shown.
@@ -183,7 +182,7 @@ class ArcGISRESTful:
         except KeyError as ex:
             raise ZeroMatched(resp["error"]["message"]) from ex
 
-    def oids_byfield(self, field: str, ids: Union[str, List[str]]) -> List[Tuple[str, ...]]:
+    def oids_byfield(self, field: str, ids: Union[str, List[str]]) -> Iterator[Tuple[str, ...]]:
         """Get Object IDs based on a list of field IDs.
 
         Parameters
@@ -209,7 +208,7 @@ class ArcGISRESTful:
 
         return self.oids_bysql(f"{field} IN ({fids})")
 
-    def oids_bysql(self, sql_clause: str) -> List[Tuple[str, ...]]:
+    def oids_bysql(self, sql_clause: str) -> Iterator[Tuple[str, ...]]:
         """Get feature IDs using a valid SQL 92 WHERE clause.
 
         Notes
@@ -244,7 +243,7 @@ class ArcGISRESTful:
         except KeyError as ex:
             raise ZeroMatched(resp["error"]["message"]) from ex
 
-    def partition_oids(self, oids: Union[List[int], int]) -> List[Tuple[str, ...]]:
+    def partition_oids(self, oids: Union[List[int], int]) -> Iterator[Tuple[str, ...]]:
         """Partition feature IDs based on ``self.max_nrecords``.
 
         Parameters
@@ -261,7 +260,7 @@ class ArcGISRESTful:
 
     def get_features(
         self,
-        featureids: List[Tuple[str, ...]],
+        featureids: Iterator[Tuple[str, ...]],
         return_m: bool = False,
         return_geom: bool = True,
     ) -> List[Dict[str, Any]]:
@@ -304,7 +303,7 @@ class WMS(WMSBase):
         string to get a list of available output formats.
     crs : str, optional
         The spatial reference system to be used for requesting the data, defaults to
-        epsg:4326.
+        ``epsg:4326``.
     version : str, optional
         The WMS service version which should be either 1.1.1 or 1.3.0, defaults to 1.3.0.
     validation : bool, optional
@@ -319,7 +318,7 @@ class WMS(WMSBase):
 
     def __init__(
         self,
-        url: AnyHttpUrl,
+        url: str,
         layers: Union[str, List[str]],
         outformat: str,
         version: str = "1.3.0",
@@ -346,7 +345,7 @@ class WMS(WMSBase):
         self,
         bbox: Tuple[float, float, float, float],
         resolution: float,
-        box_crs: str = DEF_CRS,
+        box_crs: Union[str, pyproj.CRS] = DEF_CRS,
         always_xy: bool = False,
         max_px: int = 8000000,
         kwargs: Optional[Dict[str, Any]] = None,
@@ -360,9 +359,9 @@ class WMS(WMSBase):
         resolution : float
             The output resolution in meters. The width and height of output are computed in pixel
             based on the geometry bounds and the given resolution.
-        box_crs : str, optional
+        box_crs : str, or pyproj.CRS, optional
             The spatial reference system of the input bbox, defaults to
-            epsg:4326.
+            ``epsg:4326``.
         always_xy : bool, optional
             Whether to always use xy axis order, defaults to False. Some services change the axis
             order from xy to yx, following the latest WFS version specifications but some don't.
@@ -402,15 +401,13 @@ class WMS(WMSBase):
         else:
             payload["crs"] = self.crs
 
-        geographic_crs = pyproj.CRS.from_user_input(self.crs).is_geographic
-
         def _get_payloads(
             args: Tuple[str, Tuple[Tuple[float, float, float, float], str, int, int]]
         ) -> Tuple[str, Dict[str, str]]:
             lyr, bnds = args
             _bbox, counter, _width, _height = bnds
 
-            if self.version != "1.1.1" and geographic_crs and not always_xy:
+            if self.version != "1.1.1" and pyproj.CRS(self.crs).is_geographic and not always_xy:
                 _bbox = (_bbox[1], _bbox[0], _bbox[3], _bbox[2])
             _payload = payload.copy()
             _payload["bbox"] = f'{",".join(str(c) for c in _bbox)}'
@@ -420,15 +417,14 @@ class WMS(WMSBase):
             return f"{lyr}_dd_{counter}", _payload
 
         layers, payloads = zip(*(_get_payloads(i) for i in itertools.product(self.layers, bounds)))
-        rbinary = ar.retrieve(
+        rbinary = ar.retrieve_binary(
             [self.url] * len(payloads),
-            "binary",
             [{"params": p} for p in payloads],
             max_workers=4,
             expire_after=self.expire_after,
             disable=self.disable_caching,
         )
-        return dict(zip(layers, rbinary))  # type: ignore
+        return dict(zip(layers, rbinary))
 
 
 class WFS(WFSBase):
@@ -450,7 +446,7 @@ class WFS(WFSBase):
         Defaults to 2.0.0.
     crs: str, optional
         The spatial reference system to be used for requesting the data, defaults to
-        epsg:4326.
+        ``epsg:4326``.
     read_method : str, optional
         Method for reading the retrieved data, defaults to ``json``. Valid options are
         ``json``, ``binary``, and ``text``.
@@ -470,7 +466,7 @@ class WFS(WFSBase):
 
     def __init__(
         self,
-        url: AnyHttpUrl,
+        url: str,
         layer: Optional[str] = None,
         outformat: Optional[str] = None,
         version: str = "2.0.0",
@@ -499,7 +495,7 @@ class WFS(WFSBase):
     def getfeature_bybox(
         self,
         bbox: Tuple[float, float, float, float],
-        box_crs: str = DEF_CRS,
+        box_crs: Union[str, pyproj.CRS] = DEF_CRS,
         always_xy: bool = False,
     ) -> Union[str, bytes, Dict[str, Any]]:
         """Get data from a WFS service within a bounding box.
@@ -508,9 +504,9 @@ class WFS(WFSBase):
         ----------
         bbox : tuple
             A bounding box for getting the data: [west, south, east, north]
-        box_crs : str, optional
+        box_crs : str, or pyproj.CRS, optional
             The spatial reference system of the input bbox, defaults to
-            epsg:4326.
+            ``epsg:4326``.
         always_xy : bool, optional
             Whether to always use xy axis order, defaults to False. Some services change the axis
             order from xy to yx, following the latest WFS version specifications but some don't.
@@ -523,12 +519,9 @@ class WFS(WFSBase):
             WFS query response within a bounding box.
         """
         utils.check_bbox(bbox)
+        box_crs = pyproj.CRS(box_crs)
 
-        if (
-            self.version != "1.0.0"
-            and pyproj.CRS.from_user_input(box_crs).is_geographic
-            and not always_xy
-        ):
+        if self.version != "1.0.0" and box_crs.is_geographic and not always_xy:
             bbox = (bbox[1], bbox[0], bbox[3], bbox[2])
 
         payload = {
@@ -537,7 +530,7 @@ class WFS(WFSBase):
             "outputFormat": self.outformat,
             "request": "GetFeature",
             "typeName": self.layer,
-            "bbox": f'{",".join(str(c) for c in bbox)},{box_crs}',
+            "bbox": f'{",".join(str(c) for c in bbox)},{box_crs.to_string()}',
             "srsName": self.crs,
         }
 
@@ -552,7 +545,7 @@ class WFS(WFSBase):
     def getfeature_bygeom(
         self,
         geometry: Union[Polygon, MultiPolygon],
-        geo_crs: str = DEF_CRS,
+        geo_crs: Union[str, pyproj.CRS] = DEF_CRS,
         always_xy: bool = False,
         predicate: str = "INTERSECTS",
     ) -> Union[str, bytes, Dict[str, Any]]:
@@ -562,8 +555,8 @@ class WFS(WFSBase):
         ----------
         geometry : shapely.geometry
             The input geometry
-        geo_crs : str, optional
-            The CRS of the input geometry, default to epsg:4326.
+        geo_crs : str, or pyproj.CRS, optional
+            The CRS of the input geometry, default to ``epsg:4326``.
         always_xy : bool, optional
             Whether to always use xy axis order, defaults to False. Some services change the axis
             order from xy to yx, following the latest WFS version specifications but some don't.
@@ -591,11 +584,7 @@ class WFS(WFSBase):
         """
         geom = utils.match_crs(geometry, geo_crs, self.crs)
 
-        if (
-            self.version != "1.0.0"
-            and pyproj.CRS.from_user_input(geo_crs).is_geographic
-            and not always_xy
-        ):
+        if self.version != "1.0.0" and pyproj.CRS(geo_crs).is_geographic and not always_xy:
             g_wkt = ops.transform(lambda x, y: (y, x), geom).wkt
         else:
             g_wkt = geom.wkt
