@@ -1,7 +1,6 @@
 """Tests for PyGeoOGC package."""
 import io
 import sys
-from pathlib import Path
 
 import pandas as pd
 import pytest
@@ -150,9 +149,14 @@ class TestWMS:
         """WMS by bounding box"""
         wms = WMS(self.wms_url, layers=self.layer, outformat="image/tiff", crs=DEF_CRS)
         print(wms)
-        r_dict = wms.getmap_bybox(GEO_NAT.bounds, 20, DEF_CRS, max_px=3e6)
+        r_dict = wms.getmap_bybox(GEO_NAT.bounds, 20, DEF_CRS, max_px=int(3e6))
         assert sum(sys.getsizeof(r) for r in r_dict.values()) == 11495526
 
+    def test_valid_crs(self):
+        """Get WMS valid CRSs"""
+        crs = utils.valid_wms_crs(self.wms_url)
+        assert sorted(crs) == ['epsg:3395', 'epsg:3857', 'epsg:4326']
+
 
 class TestWFS:
     wfs: WFS = WFS(
